"""Utils to perform configuration checks."""
import os

import torch
from allennlp.common import checks as allen_checks


def file_exists(*paths):
    """Check whether paths exists."""
    for path in paths:
        if path is None:
            raise allen_checks.ConfigurationError("File cannot be None")
        if not os.path.exists(path):
            raise allen_checks.ConfigurationError(f"Could not find the file at path: '{path}'")


def check_size_match(size_1: torch.Size, size_2: torch.Size, tensor_1_name: str, tensor_2_name: str):
    """Check if tensors' sizes are the same."""
    if size_1 != size_2:
        raise allen_checks.ConfigurationError(
            f"{tensor_1_name} must match {tensor_2_name}, but got {size_1} "
            f"and {size_2} instead"
        )



from .pretrained_transformer_mismatched_indexer import PretrainedTransformerMismatchedIndexer
from .token_characters_indexer import TokenCharactersIndexer
from .token_features_indexer import TokenFeatsIndexer



import torch.optim.lr_scheduler as lr_scheduler
from allennlp.training.learning_rate_schedulers import learning_rate_scheduler
from overrides import overrides


@learning_rate_scheduler.LearningRateScheduler.register("combo_scheduler")
class Scheduler(learning_rate_scheduler._PyTorchLearningRateSchedulerWrapper):

    def __init__(self, optimizer, patience: int = 6, decreases: int = 2, threshold: float = 1e-3):
        super().__init__(lr_scheduler.LambdaLR(optimizer, lr_lambda=[self._lr_lambda]))
        self.threshold = threshold
        self.decreases = decreases
        self.patience = patience
        self.start_patience = patience
        self.best_score = 0.0

    @staticmethod
    def _lr_lambda(idx: int) -> float:
        return 1.0 / (1.0 + idx * 1e-4)

    @overrides
    def step(self, metric: float = None) -> None:
        self.lr_scheduler.step()

        if metric is not None:
            if metric - self.best_score > self.threshold:
                self.best_score = metric if metric > self.best_score else self.best_score
                self.patience = self.start_patience
            else:
                if self.patience <= 1:
                    if self.decreases == 0:
                        # This is condition for Trainer to trigger early stopping
                        self.patience = 0
                    else:
                        self.patience = self.start_patience
                        self.decreases -= 1
                        self.threshold /= 2
                        self.lr_scheduler.base_lrs = [x / 2 for x in self.lr_scheduler.base_lrs]
                else:
                    self.patience -= 1



import logging
import os
from typing import List, Union, Tuple

import conllu
from allennlp import data as allen_data, common, models
from allennlp.common import util
from allennlp.data import tokenizers
from allennlp.predictors import predictor
from overrides import overrides

from unidic_combo import data
from unidic_combo.data import sentence2conllu, tokens2conllu, conllu2sentence
from unidic_combo.utils import download

logger = logging.getLogger(__name__)


@predictor.Predictor.register("semantic-multitask-predictor")
@predictor.Predictor.register("semantic-multitask-predictor-spacy", constructor="with_spacy_tokenizer")
class SemanticMultitaskPredictor(predictor.Predictor):

    def __init__(self,
                 model: models.Model,
                 dataset_reader: allen_data.DatasetReader,
                 tokenizer: allen_data.Tokenizer = tokenizers.WhitespaceTokenizer(),
                 batch_size: int = 500,
                 line_to_conllu: bool = False) -> None:
        super().__init__(model, dataset_reader)
        self.batch_size = batch_size
        self.vocab = model.vocab
        self._dataset_reader.generate_labels = False
        self._dataset_reader.lazy = True
        self._tokenizer = tokenizer
        self.without_sentence_embedding = False
        self.line_to_conllu = line_to_conllu

    def __call__(self, sentence: Union[str, List[str], List[List[str]], List[data.Sentence]]):
        """Depending on the input uses (or ignores) tokenizer.
        When model isn't only text-based only List[data.Sentence] is possible input.

        * str - tokenizer is used
        * List[str] - tokenizer is used for each string (treated as list of raw sentences)
        * List[List[str]] - tokenizer isn't used (treated as list of tokenized sentences)
        * List[data.Sentence] - tokenizer isn't used (treated as list of tokenized sentences)

        :param sentence: sentence(s) representation
        :return: Sentence or List[Sentence] depending on the input
        """
        return self.predict(sentence)

    def predict(self, sentence: Union[str, List[str], List[List[str]], List[data.Sentence]]):
        if isinstance(sentence, str):
            return data.Sentence.from_dict(self.predict_json({"sentence": sentence}))
        elif isinstance(sentence, list):
            if len(sentence) == 0:
                return []
            example = sentence[0]
            if isinstance(example, str) or isinstance(example, list):
                sentences = []
                for sentences_batch in util.lazy_groups_of(sentence, self.batch_size):
                    sentences_batch = self.predict_batch_json([self._to_input_json(s) for s in sentences_batch])
                    sentences.extend(sentences_batch)
                return sentences
            elif isinstance(example, data.Sentence):
                sentences = []
                for sentences_batch in util.lazy_groups_of(sentence, self.batch_size):
                    sentences_batch = self.predict_batch_instance([self._to_input_instance(s) for s in sentences_batch])
                    sentences.extend(sentences_batch)
                return sentences
            else:
                raise ValueError("List must have either sentences as str, List[str] or Sentence object.")
        else:
            raise ValueError("Input must be either string or list of strings.")

    @overrides
    def predict_batch_instance(self, instances: List[allen_data.Instance]) -> List[data.Sentence]:
        sentences = []
        predictions = super().predict_batch_instance(instances)
        for prediction, instance in zip(predictions, instances):
            tree, sentence_embedding = self.predict_instance_as_tree(instance)
            sentence = conllu2sentence(tree, sentence_embedding)
            sentences.append(sentence)
        return sentences

    @overrides
    def predict_batch_json(self, inputs: List[common.JsonDict]) -> List[data.Sentence]:
        sentences = []
        instances = self._batch_json_to_instances(inputs)
        predictions = self.predict_batch_instance(instances)
        for prediction, instance in zip(predictions, instances):
            tree, sentence_embedding = self.predict_instance_as_tree(instance)
            sentence = conllu2sentence(tree, sentence_embedding)
            sentences.append(sentence)
        return sentences

    @overrides
    def predict_instance(self, instance: allen_data.Instance, serialize: bool = True) -> data.Sentence:
        tree, sentence_embedding = self.predict_instance_as_tree(instance)
        return conllu2sentence(tree, sentence_embedding)

    @overrides
    def predict_json(self, inputs: common.JsonDict) -> data.Sentence:
        instance = self._json_to_instance(inputs)
        tree, sentence_embedding = self.predict_instance_as_tree(instance)
        return conllu2sentence(tree, sentence_embedding)

    def predict_instance_as_tree(self, instance: allen_data.Instance) -> Tuple[conllu.TokenList, List[float]]:
        predictions = super().predict_instance(instance)
        return self._predictions_as_tree(predictions, instance)

    @overrides
    def _json_to_instance(self, json_dict: common.JsonDict) -> allen_data.Instance:
        sentence = json_dict["sentence"]
        if isinstance(sentence, str):
            tokens = [t.text for t in self._tokenizer.tokenize(json_dict["sentence"])]
        elif isinstance(sentence, list):
            tokens = sentence
        else:
            raise ValueError("Input must be either string or list of strings.")
        return self._dataset_reader.text_to_instance(tokens2conllu(tokens))

    @overrides
    def load_line(self, line: str) -> common.JsonDict:
        return self._to_input_json(line.replace("\n", "").strip())

    @overrides
    def dump_line(self, outputs: data.Sentence) -> str:
        # Check whether serialized (str) tree or token's list
        # Serialized tree has already separators between lines
        if self.without_sentence_embedding:
            outputs.sentence_embedding = []
        if self.line_to_conllu:
            return sentence2conllu(outputs, keep_semrel=self._dataset_reader.use_sem).serialize()
        else:
            return outputs.to_json()

    @staticmethod
    def _to_input_json(sentence: str):
        return {"sentence": sentence}

    def _to_input_instance(self, sentence: data.Sentence) -> allen_data.Instance:
        return self._dataset_reader.text_to_instance(sentence2conllu(sentence))

    def _predictions_as_tree(self, predictions, instance):
        tree = instance.fields["metadata"]["input"]
        field_names = instance.fields["metadata"]["field_names"]
        tree_tokens = [t for t in tree if isinstance(t["id"], int)]
        for idx, token in enumerate(tree_tokens):
            for field_name in field_names:
                if field_name in predictions:
                    if field_name in ["xpostag", "upostag", "semrel", "deprel"]:
                        value = self.vocab.get_token_from_index(predictions[field_name][idx], field_name + "_labels")
                        token[field_name] = value
                    elif field_name in ["head"]:
                        token[field_name] = int(predictions[field_name][idx])
                    elif field_name in ["feats"]:
                        slices = self._model.morphological_feat.slices
                        features = []
                        prediction = predictions[field_name][idx]
                        for (cat, cat_indices), pred_idx in zip(slices.items(), prediction):
                            if cat not in ["__PAD__", "_"]:
                                value = self.vocab.get_token_from_index(cat_indices[pred_idx],
                                                                        field_name + "_labels")
                                # Exclude auxiliary values
                                if "=None" not in value:
                                    features.append(value)
                        if len(features) == 0:
                            field_value = "_"
                        else:
                            field_value = "|".join(sorted(features))

                        token[field_name] = field_value
                    elif field_name == "head":
                        pass
                    elif field_name == "lemma":
                        prediction = predictions[field_name][idx]
                        word_chars = []
                        for char_idx in prediction[1:-1]:
                            pred_char = self.vocab.get_token_from_index(char_idx, "lemma_characters")

                            if pred_char == "__END__":
                                break
                            elif pred_char == "__PAD__":
                                continue
                            elif "_" in pred_char:
                                pred_char = "?"

                            word_chars.append(pred_char)
                        token[field_name] = "".join(word_chars)
                    else:
                        raise NotImplementedError(f"Unknown field name {field_name}!")

        return tree, predictions["sentence_embedding"]

    @classmethod
    def with_spacy_tokenizer(cls, model: models.Model,
                             dataset_reader: allen_data.DatasetReader):
        return cls(model, dataset_reader, tokenizers.SpacyTokenizer())

    @classmethod
    def from_pretrained(cls, path: str, tokenizer=tokenizers.SpacyTokenizer(),
                        batch_size: int = 500,
                        cuda_device: int = -1):
        util.import_module_and_submodules("unidic_combo.commands")
        util.import_module_and_submodules("unidic_combo.models")
        util.import_module_and_submodules("unidic_combo.training")

        if os.path.exists(path):
            model_path = path
        else:
            try:
                logger.debug("Downloading model.")
                model_path = download.download_file(path)
            except Exception as e:
                logger.error(e)
                raise e

        archive = models.load_archive(model_path, cuda_device=cuda_device)
        model = archive.model
        dataset_reader = allen_data.DatasetReader.from_params(
            archive.config["dataset_reader"])
        return cls(model, dataset_reader, tokenizer, batch_size)



import collections
from dataclasses import dataclass, field
from typing import Optional, List, Dict, Any, Union, Tuple

import conllu
from dataclasses_json import dataclass_json
from overrides import overrides


@dataclass_json
@dataclass
class Token:
    id: Optional[Union[int, Tuple]] = None
    token: Optional[str] = None
    lemma: Optional[str] = None
    upostag: Optional[str] = None
    xpostag: Optional[str] = None
    feats: Optional[str] = None
    head: Optional[int] = None
    deprel: Optional[str] = None
    deps: Optional[str] = None
    misc: Optional[str] = None
    semrel: Optional[str] = None


@dataclass_json
@dataclass
class Sentence:
    tokens: List[Token] = field(default_factory=list)
    sentence_embedding: List[float] = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=collections.OrderedDict)


class _TokenList(conllu.TokenList):

    @overrides
    def __repr__(self):
        return 'TokenList<' + ', '.join(token['token'] for token in self) + '>'


def sentence2conllu(sentence: Sentence, keep_semrel: bool = True) -> conllu.TokenList:
    tokens = []
    for token in sentence.tokens:
        token_dict = collections.OrderedDict(token.to_dict())
        # Remove semrel to have default conllu format.
        if not keep_semrel:
            del token_dict["semrel"]
        tokens.append(token_dict)
    # Range tokens must be tuple not list, this is conllu library requirement
    for t in tokens:
        if type(t["id"]) == list:
            t["id"] = tuple(t["id"])
    return _TokenList(tokens=tokens,
                      metadata=sentence.metadata)


def tokens2conllu(tokens: List[str]) -> conllu.TokenList:
    return _TokenList(
        [collections.OrderedDict({"id": idx, "token": token}) for
         idx, token
         in enumerate(tokens, start=1)],
        metadata=collections.OrderedDict()
    )


def conllu2sentence(conllu_sentence: conllu.TokenList,
                    sentence_embedding: List[float]) -> Sentence:
    return Sentence(
        tokens=[Token.from_dict(t) for t in conllu_sentence.tokens],
        sentence_embedding=sentence_embedding,
        metadata=conllu_sentence.metadata
    )



"""Sequence multilabel field implementation."""
import logging
import textwrap
from typing import Set, List, Callable, Iterator, Union, Dict

import torch
from allennlp import data
from allennlp.common import checks, util
from allennlp.data import fields
from overrides import overrides

logger = logging.getLogger(__name__)


class SequenceMultiLabelField(data.Field[torch.Tensor]):
    """
    A `SequenceMultiLabelField` is an extension of the :class:`MultiLabelField` that allows for multiple labels
    while keeping sequence dimension.

    This field will get converted into a sequence of vectors of length equal to the vocabulary size with
    M from N encoding for the labels (all zeros, and ones for the labels).

    # Parameters

    multi_labels : `List[List[str]]`
    multi_label_indexer : `Callable[[data.Vocabulary], Callable[[List[str]], List[int]]]`
        Nested callable which based on vocab creates mapper for multilabel field in the sequence from strings
        to indexed, int values.
    sequence_field : `SequenceField`
        A field containing the sequence that this `SequenceMultiLabelField` is labeling.  Most often, this is a
        `TextField`, for tagging individual tokens in a sentence.
    label_namespace : `str`, optional (default="labels")
        The namespace to use for converting label strings into integers.  We map label strings to
        integers for you (e.g., "entailment" and "contradiction" get converted to 0, 1, ...),
        and this namespace tells the `Vocabulary` object which mapping from strings to integers
        to use (so "entailment" as a label doesn't get the same integer id as "entailment" as a
        word).  If you have multiple different label fields in your data, you should make sure you
        use different namespaces for each one, always using the suffix "labels" (e.g.,
        "passage_labels" and "question_labels").
    """
    _already_warned_namespaces: Set[str] = set()

    def __init__(
            self,
            multi_labels: List[List[str]],
            multi_label_indexer: Callable[[data.Vocabulary], Callable[[List[str]], List[int]]],
            sequence_field: fields.SequenceField,
            label_namespace: str = "labels",
    ) -> None:
        self.multi_labels = multi_labels
        self.sequence_field = sequence_field
        self.multi_label_indexer = multi_label_indexer
        self._label_namespace = label_namespace
        self._indexed_multi_labels = None
        self._maybe_warn_for_namespace(label_namespace)
        if len(multi_labels) != sequence_field.sequence_length():
            raise checks.ConfigurationError(
                "Label length and sequence length "
                "don't match: %d and %d" % (len(multi_labels), sequence_field.sequence_length())
            )

        if not all([isinstance(x, str) for multi_label in multi_labels for x in multi_label]):
            raise checks.ConfigurationError(
                "SequenceMultiLabelField must be passed either all "
                "strings or all ints. Found labels {} with "
                "types: {}.".format(multi_labels, [type(x) for multi_label in multi_labels for x in multi_label])
            )

    def _maybe_warn_for_namespace(self, label_namespace: str) -> None:
        if not (self._label_namespace.endswith("labels") or self._label_namespace.endswith("tags")):
            if label_namespace not in self._already_warned_namespaces:
                logger.warning(
                    "Your label namespace was '%s'. We recommend you use a namespace "
                    "ending with 'labels' or 'tags', so we don't add UNK and PAD tokens by "
                    "default to your vocabulary.  See documentation for "
                    "`non_padded_namespaces` parameter in Vocabulary.",
                    self._label_namespace,
                )
                self._already_warned_namespaces.add(label_namespace)

    # Sequence methods
    def __iter__(self) -> Iterator[Union[List[str], int]]:
        return iter(self.multi_labels)

    def __getitem__(self, idx: int) -> Union[List[str], int]:
        return self.multi_labels[idx]

    def __len__(self) -> int:
        return len(self.multi_labels)

    @overrides
    def count_vocab_items(self, counter: Dict[str, Dict[str, int]]):
        if self._indexed_multi_labels is None:
            for multi_label in self.multi_labels:
                for label in multi_label:
                    counter[self._label_namespace][label] += 1  # type: ignore

    @overrides
    def index(self, vocab: data.Vocabulary):
        indexer = self.multi_label_indexer(vocab)

        indexed = []
        for multi_label in self.multi_labels:
            indexed.append(indexer(multi_label))
        self._indexed_multi_labels = indexed

    @overrides
    def get_padding_lengths(self) -> Dict[str, int]:
        return {"num_tokens": self.sequence_field.sequence_length()}

    @overrides
    def as_tensor(self, padding_lengths: Dict[str, int]) -> torch.Tensor:
        desired_num_tokens = padding_lengths["num_tokens"]
        assert len(self._indexed_multi_labels) > 0
        classes_count = len(self._indexed_multi_labels[0])
        default_value = [0.0] * classes_count
        padded_tags = util.pad_sequence_to_length(self._indexed_multi_labels, desired_num_tokens, lambda: default_value)
        tensor = torch.LongTensor(padded_tags)
        return tensor

    @overrides
    def empty_field(self) -> "SequenceMultiLabelField":
        # The empty_list here is needed for mypy
        empty_list: List[List[str]] = [[]]
        sequence_label_field = SequenceMultiLabelField(empty_list, lambda x: lambda y: y,
                                                       self.sequence_field.empty_field())
        sequence_label_field._indexed_labels = empty_list
        return sequence_label_field

    def __str__(self) -> str:
        length = self.sequence_field.sequence_length()
        formatted_labels = "".join(
            "\t\t" + labels + "\n" for labels in textwrap.wrap(repr(self.multi_labels), 100)
        )
        return (
            f"SequenceMultiLabelField of length {length} with "
            f"labels:\n {formatted_labels} \t\tin namespace: '{self._label_namespace}'."
        )



from typing import Union, Any, Dict, Tuple

from allennlp import training
from allennlp.training import trainer as allen_trainer


@training.Checkpointer.register("finishing_only_checkpointer")
class FinishingTrainingCheckpointer(training.Checkpointer):
    """Checkpointer disables restoring interrupted training and saves only weights
       when this is last epoch / learning rate is on the last lr decrease.

       Remove checkpointer configuration from config template to get regular, on best score, saving."""

    def save_checkpoint(
            self,
            epoch: Union[int, str],
            trainer: "allen_trainer.Trainer",
            is_best_so_far: bool = False,
    ) -> None:
        if trainer._learning_rate_scheduler.decreases <= 1 or epoch == trainer._num_epochs - 1:
            super().save_checkpoint(epoch, trainer, is_best_so_far)

    def restore_checkpoint(self) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        return {}, {}

    def maybe_save_checkpoint(
            self, trainer: "allen_trainer.Trainer", epoch: int, batches_this_epoch: int
    ) -> None:
        pass



"""Encoder."""
from typing import Optional, Tuple

import torch
import torch.nn.utils.rnn as rnn
from allennlp import common, modules
from allennlp.modules import input_variational_dropout, stacked_bidirectional_lstm, seq2seq_encoders
from overrides import overrides


class StackedBiLSTM(stacked_bidirectional_lstm.StackedBidirectionalLstm, common.FromParams):

    def __init__(self, input_size: int, hidden_size: int, num_layers: int, recurrent_dropout_probability: float,
                 layer_dropout_probability: float, use_highway: bool = False):
        super().__init__(input_size=input_size,
                         hidden_size=hidden_size,
                         num_layers=num_layers,
                         recurrent_dropout_probability=recurrent_dropout_probability,
                         layer_dropout_probability=layer_dropout_probability,
                         use_highway=use_highway)

    @overrides
    def forward(self,
                inputs: rnn.PackedSequence,
                initial_state: Optional[Tuple[torch.Tensor, torch.Tensor]] = None
                ) -> Tuple[rnn.PackedSequence, Tuple[torch.Tensor, torch.Tensor]]:
        """Changes when compared to stacked_bidirectional_lstm.StackedBidirectionalLstm
        * dropout also on last layer
        * accepts BxTxD tensor
        * state from n-1 layer used as n layer initial state

        :param inputs:
        :param initial_state:
        :return:
        """
        output_sequence = inputs
        state_fwd = None
        state_bwd = None
        for i in range(self.num_layers):
            forward_layer = getattr(self, f"forward_layer_{i}")
            backward_layer = getattr(self, f"backward_layer_{i}")

            forward_output, state_fwd = forward_layer(output_sequence, state_fwd)
            backward_output, state_bwd = backward_layer(output_sequence, state_bwd)

            forward_output, lengths = rnn.pad_packed_sequence(forward_output, batch_first=True)
            backward_output, _ = rnn.pad_packed_sequence(backward_output, batch_first=True)

            output_sequence = torch.cat([forward_output, backward_output], -1)

            output_sequence = self.layer_dropout(output_sequence)
            output_sequence = rnn.pack_padded_sequence(output_sequence, lengths, batch_first=True)

        return output_sequence, (state_fwd, state_bwd)


@modules.Seq2SeqEncoder.register("combo_encoder")
class ComboEncoder(seq2seq_encoders.PytorchSeq2SeqWrapper):
    """COMBO encoder (https://www.aclweb.org/anthology/K18-2004.pdf).

    This implementation uses Variational Dropout on the input and then outputs of each BiLSTM layer
    (instead of used Gaussian Dropout and Gaussian Noise).
    """

    def __init__(self, stacked_bilstm: StackedBiLSTM, layer_dropout_probability: float):
        super().__init__(stacked_bilstm, stateful=False)
        self.layer_dropout = input_variational_dropout.InputVariationalDropout(p=layer_dropout_probability)

    @overrides
    def forward(self,
                inputs: torch.Tensor,
                mask: torch.BoolTensor,
                hidden_state: torch.Tensor = None) -> torch.Tensor:
        x = self.layer_dropout(inputs)
        x = super().forward(x, mask)
        return self.layer_dropout(x)
